from .nameko_cachetools import CachedRpcProxy, CacheFirstRpcProxy

__author__ = 'Santiago Suarez Ordonez'
__email__ = 'santiycr@gmail.com'
__version__ = '1.0.0'
__all__ = ['CachedRpcProxy', 'CacheFirstRpcProxy']
